import tempfile

import pytest
import responses

from instabot import utils
from instabot.api.config import API_URL, SIG_KEY_VERSION

from .test_bot import TestBot
from .test_variables import (
    TEST_CAPTION_ITEM,
    TEST_COMMENT_ITEM,
    TEST_COMMENT_LIKER_ITEM,
    TEST_FOLLOWER_ITEM,
    TEST_FOLLOWING_ITEM,
    TEST_INBOX_THREAD_ITEM,
    TEST_LOCATION_ITEM,
    TEST_MEDIA_LIKER,
    TEST_MOST_RECENT_INVITER_ITEM,
    TEST_PHOTO_ITEM,
    TEST_SEARCH_USERNAME_ITEM,
    TEST_TIMELINE_PHOTO_ITEM,
    TEST_USER_ITEM,
    TEST_USER_TAG_ITEM,
    TEST_USERNAME_INFO_ITEM,
)

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


class TestBotGet(TestBot):
    @responses.activate
    def test_get_media_owner(self):
        media_id = 1234

        responses.add(
            responses.GET,
            "{api_url}media/{media_id}/info/".format(
                api_url=API_URL, media_id=media_id
            ),
            json={
                "auto_load_more_enabled": True,
                "num_results": 1,
                "status": "ok",
                "more_available": False,
                "items": [TEST_PHOTO_ITEM],
            },
            status=200,
        )
        # responses.add(
        #     responses.POST, "{api_url}media/{media_id}/info/".format(
        #     api_url=API_URL, media_id=media_id),
        #     json={"status": "ok"}, status=200)

        owner = self.bot.get_media_owner(media_id)

        assert owner == str(TEST_PHOTO_ITEM["user"]["pk"])

        # owner = self.bot.get_media_owner(media_id)

        # assert owner is False

    @responses.activate
    def test_get_media_info(self):
        media_id = 1234

        responses.add(
            responses.GET,
            "{api_url}media/{media_id}/info/".format(
                api_url=API_URL, media_id=media_id
            ),
            json={
                "auto_load_more_enabled": True,
                "num_results": 1,
                "status": "ok",
                "more_available": False,
                "items": [TEST_PHOTO_ITEM],
            },
            status=200,
        )
        # responses.add(
        #     responses.POST, "{api_url}media/{media_id}/info/".format(
        #     api_url=API_URL, media_id=media_id),
        #     json={"status": "ok"}, status=200)

        expected_result = {}
        for key in TEST_PHOTO_ITEM:
            expected_result[key] = TEST_PHOTO_ITEM[key]

        result = self.bot.get_media_info(media_id)

        assert result[0] == expected_result

    @responses.activate
    def test_get_popular_medias(self):
        results = 5
        responses.add(
            responses.GET,
            (
                "{api_url}feed/popular/?people_teaser_supported=1" +
                "&rank_token={rank_token}&ranked_content=true&"
            ).format(
                api_url=API_URL, rank_token=self.bot.api.rank_token
            ),
            json={
                "auto_load_more_enabled": True,
                "num_results": results,
                "status": "ok",
                "more_available": False,
                "items": [TEST_PHOTO_ITEM for _ in range(results)],
            },
            status=200,
        )

        medias = self.bot.get_popular_medias()

        assert medias == [str(TEST_PHOTO_ITEM["id"]) for _ in range(results)]
        assert len(medias) == results

    @responses.activate
    def test_get_timeline_medias(self):
        self.bot.max_likes_to_like = TEST_PHOTO_ITEM["like_count"] + 1
        results = 8
        responses.add(
            responses.POST,
            "{api_url}feed/timeline/".format(api_url=API_URL),
            json={
                "auto_load_more_enabled": True,
                "num_results": results,
                "is_direct_v2_enabled": True,
                "status": "ok",
                "next_max_id": None,
                "more_available": False,
                "feed_items": [
                    TEST_TIMELINE_PHOTO_ITEM for _ in range(results)
                ],
            },
            status=200,
        )
        responses.add(
            responses.POST,
            "{api_url}feed/timeline/".format(api_url=API_URL),
            json={"status": "fail"},
            status=400,
        )

        medias = self.bot.get_timeline_medias()

        assert medias == [TEST_PHOTO_ITEM["id"] for _ in range(results)]
        assert len(medias) == results

        medias = self.bot.get_timeline_medias()

        assert medias == []
        assert len(medias) == 0

    @responses.activate
    def test_get_timeline_users(self):
        results = 8
        responses.add(
            responses.POST,
            "{api_url}feed/timeline/".format(api_url=API_URL),
            json={
                "auto_load_more_enabled": True,
                "num_results": results,
                "is_direct_v2_enabled": True,
                "status": "ok",
                "next_max_id": None,
                "more_available": False,
                "feed_items": [
                    TEST_TIMELINE_PHOTO_ITEM for _ in range(results)
                ],
            },
            status=200,
        )
        responses.add(
            responses.POST,
            "{api_url}feed/timeline/".format(api_url=API_URL),
            json={"status": "fail"},
            status=400,
        )

        users = self.bot.get_timeline_users()

        assert users == [
            str(TEST_TIMELINE_PHOTO_ITEM["media_or_ad"]["user"]["pk"])
            for _ in range(results)
        ]
        assert len(users) == results

        users = self.bot.get_timeline_users()

        assert users == []
        assert len(users) == 0

    @responses.activate
    def test_get_your_medias(self):
        results = 5
        my_test_photo_item = TEST_PHOTO_ITEM.copy()
        my_test_photo_item["user"]["pk"] = self.USER_ID
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results,
            "status": "ok",
            "more_available": False,
            "items": [my_test_photo_item for _ in range(results)],
        }
        responses.add(
            responses.GET,
            "{api_url}feed/user/{user_id}/".format(
                api_url=API_URL,
                user_id=self.bot.user_id,
                rank_token=self.bot.api.rank_token,
            ),
            json=response_data,
            status=200,
        )

        medias = self.bot.get_your_medias()

        assert medias == [my_test_photo_item["id"] for _ in range(results)]
        assert len(medias) == results

        medias = self.bot.get_your_medias(as_dict=True)

        assert medias == response_data["items"]
        assert len(medias) == results

    @responses.activate
    @pytest.mark.parametrize("user_id", [1234567890, "1234567890"])
    def test_get_user_medias(self, user_id):
        results = 4
        my_test_photo_item = TEST_PHOTO_ITEM.copy()
        my_test_photo_item["user"]["pk"] = user_id
        my_test_photo_items = []
        for _ in range(results):
            my_test_photo_items.append(my_test_photo_item.copy())
        expect_filtered = 0
        my_test_photo_items[1]["has_liked"] = True
        expect_filtered += 1
        my_test_photo_items[2]["like_count"] = self.bot.max_likes_to_like + 1
        expect_filtered += 1
        my_test_photo_items[3]["like_count"] = self.bot.max_likes_to_like - 1
        expect_filtered += 1
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results,
            "status": "ok",
            "more_available": False,
            "items": my_test_photo_items,
        }
        responses.add(
            responses.GET,
            "{api_url}feed/user/{user_id}/".format(
                api_url=API_URL,
                user_id=user_id,
                rank_token=self.bot.api.rank_token
            ),
            json=response_data,
            status=200,
        )

        # no need to test is_comment=True because there's no item 'comments' in
        # user feed object returned by `feed/user/{user_id}/` API call.

        medias = self.bot.get_user_medias(
            user_id,
            filtration=False,
            is_comment=False
        )
        assert medias == [
            test_photo_item["id"] for test_photo_item in my_test_photo_items
        ]
        assert len(medias) == results

        medias = self.bot.get_user_medias(
            user_id,
            filtration=True,
            is_comment=False
        )
        assert medias == [
            test_photo_item["id"]
            for test_photo_item in my_test_photo_items
            if (
                not test_photo_item["has_liked"]
                and test_photo_item["like_count"] < self.bot.max_likes_to_like
                and test_photo_item["like_count"] > self.bot.min_likes_to_like
            )
        ]
        assert len(medias) == results - expect_filtered

    @responses.activate
    def test_get_archived_medias(self):
        results = 5
        my_test_photo_item = TEST_PHOTO_ITEM.copy()
        my_test_photo_item["user"]["pk"] = self.USER_ID
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results,
            "status": "ok",
            "more_available": False,
            "items": [my_test_photo_item for _ in range(results)],
        }
        responses.add(
            responses.GET,
            (
                "{api_url}feed/only_me_feed/?rank_token={rank_token}&" +
                "ranked_content=true&"
            ).format(
                api_url=API_URL, rank_token=self.bot.api.rank_token
            ),
            json=response_data,
            status=200,
        )

        medias = self.bot.get_archived_medias()

        assert medias == [my_test_photo_item["id"] for _ in range(results)]
        assert len(medias) == results

        medias = self.bot.get_archived_medias(as_dict=True)

        assert medias == response_data["items"]
        assert len(medias) == results

    @responses.activate
    def test_search_users(self):
        results = 5
        query = "test"
        my_test_user_item = TEST_USER_ITEM
        response_data = {
            "has_more": True,
            "num_results": results,
            "rank_token": self.bot.api.rank_token,
            "status": "ok",
            "users": [my_test_user_item for _ in range(results)],
        }
        responses.add(
            responses.GET,
            (
                "{api_url}users/search/?ig_sig_key_version={sig_key}&" +
                "is_typeahead=true&query={query}&rank_token={rank_token}"
            ).format(
                api_url=API_URL,
                rank_token=self.bot.api.rank_token,
                query=query,
                sig_key=SIG_KEY_VERSION,
            ),
            json=response_data,
            status=200,
        )

        medias = self.bot.search_users(query)

        assert medias == [str(my_test_user_item["pk"]) for _ in range(results)]
        assert len(medias) == results

    @responses.activate
    def test_search_users_failed(self):
        query = "test"
        response_data = {"status": "fail"}
        responses.add(
            responses.GET,
            (
                "{api_url}users/search/?ig_sig_key_version={sig_key}" +
                "&is_typeahead=true&query={query}&rank_token={rank_token}"
            ).format(
                api_url=API_URL,
                rank_token=self.bot.api.rank_token,
                query=query,
                sig_key=SIG_KEY_VERSION,
            ),
            json=response_data,
            status=200,
        )

        medias = self.bot.search_users(query)

        assert medias == []

    @responses.activate
    def test_get_comments(self):
        results = 5
        response_data = {
            "caption": TEST_CAPTION_ITEM,
            "caption_is_edited": False,
            "comment_count": 4,
            "comment_likes_enabled": True,
            "comments": [TEST_COMMENT_ITEM for _ in range(results)],
            "has_more_comments": False,
            "has_more_headload_comments": False,
            "media_header_display": "none",
            "preview_comments": [],
            "status": "ok",
        }
        media_id = 1234567890
        responses.add(
            responses.GET,
            "{api_url}media/{media_id}/comments/?".format(
                api_url=API_URL, media_id=media_id
            ),
            json=response_data,
            status=200,
        )

        comments = self.bot.get_media_comments(media_id)
        assert comments == response_data["comments"]
        assert len(comments) == results

    @responses.activate
    def test_get_comments_text(self):
        results = 5
        response_data = {
            "caption": TEST_CAPTION_ITEM,
            "caption_is_edited": False,
            "comment_count": 4,
            "comment_likes_enabled": True,
            "comments": [TEST_COMMENT_ITEM for _ in range(results)],
            "has_more_comments": False,
            "has_more_headload_comments": False,
            "media_header_display": "none",
            "preview_comments": [],
            "status": "ok",
        }
        media_id = 1234567890
        responses.add(
            responses.GET,
            "{api_url}media/{media_id}/comments/?".format(
                api_url=API_URL, media_id=media_id
            ),
            json=response_data,
            status=200,
        )

        comments = self.bot.get_media_comments(media_id, only_text=True)
        expected_result = [
            comment["text"] for comment in response_data["comments"]
        ]

        assert comments == expected_result
        assert len(comments) == results

    @responses.activate
    def test_get_comments_failed(self):
        response_data = {"status": "fail"}
        media_id = 1234567890
        responses.add(
            responses.GET,
            "{api_url}media/{media_id}/comments/?".format(
                api_url=API_URL, media_id=media_id
            ),
            json=response_data,
            status=200,
        )

        comments = self.bot.get_media_comments(media_id)
        assert comments == []

    @responses.activate
    def test_get_commenters(self):
        results = 5
        response_data = {
            "caption": TEST_CAPTION_ITEM,
            "caption_is_edited": False,
            "comment_count": 4,
            "comment_likes_enabled": True,
            "comments": [TEST_COMMENT_ITEM for _ in range(results)],
            "has_more_comments": False,
            "has_more_headload_comments": False,
            "media_header_display": "none",
            "preview_comments": [],
            "status": "ok",
        }
        media_id = 1234567890
        responses.add(
            responses.GET,
            "{api_url}media/{media_id}/comments/?".format(
                api_url=API_URL, media_id=media_id
            ),
            json=response_data,
            status=200,
        )

        expected_commenters = [
            str(TEST_COMMENT_ITEM["user"]["pk"]) for _ in range(results)
        ]

        commenters = self.bot.get_media_commenters(media_id)
        assert commenters == expected_commenters
        assert len(commenters) == results

    @responses.activate
    def test_get_commenters_failed(self):
        response_data = {"status": "fail"}
        media_id = 1234567890
        responses.add(
            responses.GET,
            "{api_url}media/{media_id}/comments/?".format(
                api_url=API_URL, media_id=media_id
            ),
            json=response_data,
            status=200,
        )

        expected_commenters = []

        commenters = self.bot.get_media_commenters(media_id)
        assert commenters == expected_commenters

    @pytest.mark.parametrize(
        "url,result",
        [
            ("https://www.instagram.com/p/BfHrDvCDuzC/", 1713527555896569026),
            ("test", False),
        ],
    )
    def test_get_media_id_from_link_with_wrong_data(self, url, result):
        media_id = self.bot.get_media_id_from_link(url)

        assert result == media_id

    @responses.activate
    @pytest.mark.parametrize(
        "comments", [["comment1", "comment2", "comment3"], [], None]
    )
    def test_get_comment(self, comments):
        fname = tempfile.mkstemp()[1]  # Temporary file
        self.bot.comments_file = utils.file(fname, verbose=True)
        if comments:
            for comment in comments:
                self.bot.comments_file.append(comment)
            assert self.bot.get_comment() in self.bot.comments_file.list
        else:
            assert self.bot.get_comment() == "Wow!"

    @responses.activate
    @pytest.mark.parametrize("user_id", [1234, "1234"])
    def test_get_username_info(self, user_id):
        response_data = {"status": "ok", "user": TEST_USERNAME_INFO_ITEM}
        expected_result = {}
        for key in TEST_USERNAME_INFO_ITEM:
            expected_result[key] = TEST_USERNAME_INFO_ITEM[key]

        responses.add(
            responses.GET,
            "{api_url}users/{user_id}/info/".format(
                api_url=API_URL,
                user_id=user_id
            ),
            status=200,
            json=response_data,
        )

        result = self.bot.get_user_info(user_id)

        assert result == expected_result

    @responses.activate
    @pytest.mark.parametrize("user_id", [1234, "1234"])
    @patch("time.sleep", return_value=None)
    def test_get_username_from_user_id(self, patched_time_sleep, user_id):
        response_data = {"status": "ok", "user": TEST_USERNAME_INFO_ITEM}
        expected_user_id = str(TEST_USERNAME_INFO_ITEM["username"])

        responses.add(
            responses.GET,
            "{api_url}users/{user_id}/info/".format(
                api_url=API_URL,
                user_id=user_id
            ),
            status=200,
            json=response_data,
        )

        result = self.bot.get_username_from_user_id(user_id)

        assert result == expected_user_id

    @responses.activate
    @pytest.mark.parametrize("user_id", ["123231231231234", 123231231231234])
    @patch("time.sleep", return_value=None)
    def test_get_username_from_user_id_404(self, patched_time_sleep, user_id):
        response_data = {"status": "fail", "message": "User not found"}
        responses.add(
            responses.GET,
            "{api_url}users/{user_id}/info/".format(
                api_url=API_URL,
                user_id=user_id
            ),
            status=404,
            json=response_data,
        )

        assert not self.bot.get_username_from_user_id(user_id)

    @responses.activate
    @pytest.mark.parametrize("username", ["@test", "test", "1234"])
    @patch("time.sleep", return_value=None)
    def test_get_user_id_from_username(self, patched_time_sleep, username):
        response_data = {"status": "ok", "user": TEST_SEARCH_USERNAME_ITEM}
        expected_user_id = str(TEST_SEARCH_USERNAME_ITEM["pk"])

        responses.add(
            responses.GET,
            "{api_url}users/{username}/usernameinfo/".format(
                api_url=API_URL, username=username
            ),
            status=200,
            json=response_data,
        )

        result = self.bot.get_user_id_from_username(username)
        del self.bot._usernames[username]  # Invalidate cache

        assert result == expected_user_id

    @responses.activate
    @pytest.mark.parametrize(
        "username", ["usernotfound", "nottexisteduser", "123231231231234"]
    )
    @patch("time.sleep", return_value=None)
    def test_get_user_id_from_username_404(self, patched_time_sleep, username):
        response_data = {"status": "fail", "message": "User not found"}
        responses.add(
            responses.GET,
            "{api_url}users/{username}/usernameinfo/".format(
                api_url=API_URL, username=username
            ),
            status=404,
            json=response_data,
        )

        assert not self.bot.get_user_id_from_username(username)

    @responses.activate
    @pytest.mark.parametrize(
        "username,url,result",
        [
            ("@test", "test", str(TEST_SEARCH_USERNAME_ITEM["pk"])),
            ("test", "test", str(TEST_SEARCH_USERNAME_ITEM["pk"])),
            ("1234", "1234", "1234"),
            (1234, "1234", "1234"),
        ],
    )
    @patch("time.sleep", return_value=None)
    def test_convert_to_user_id(
        self,
        patched_time_sleep,
        username,
        url,
        result
    ):
        response_data = {"status": "ok", "user": TEST_SEARCH_USERNAME_ITEM}
        responses.add(
            responses.GET,
            "{api_url}users/{username}/usernameinfo/".format(
                api_url=API_URL, username=url
            ),
            status=200,
            json=response_data,
        )

        user_id = self.bot.convert_to_user_id(username)

        assert result == user_id

    @responses.activate
    @pytest.mark.parametrize("user_id", ["3998456661", 3998456661])
    def test_get_user_tags_medias(self, user_id):
        results = 8
        responses.add(
            responses.GET,
            (
                "{api_url}usertags/{user_id}/feed/?rank_token={rank_token}" +
                "&ranked_content=true&"
            ).format(
                api_url=API_URL,
                user_id=user_id,
                rank_token=self.bot.api.rank_token
            ),
            json={
                "status": "ok",
                "num_results": results,
                "auto_load_more_enabled": True,
                "items": [TEST_USER_TAG_ITEM for _ in range(results)],
                "more_available": False,
                "total_count": results,
                "requires_review": False,
                "new_photos": [],
            },
            status=200,
        )

        medias = self.bot.get_user_tags_medias(user_id)

        assert medias == [
            str(TEST_USER_TAG_ITEM["pk"]) for _ in range(results)
        ]
        assert len(medias) == results

    @responses.activate
    @pytest.mark.parametrize("hashtag", ["hashtag1", "hashtag2"])
    def test_get_hashtag_medias(self, hashtag):

        results = 5
        my_test_photo_item = TEST_PHOTO_ITEM.copy()
        my_test_photo_item["like_count"] = self.bot.min_likes_to_like + 1
        my_test_photo_items = []
        for _ in range(results):
            my_test_photo_items.append(my_test_photo_item.copy())
        expect_filtered = 0
        my_test_photo_items[1]["has_liked"] = True
        expect_filtered += 1
        my_test_photo_items[2]["like_count"] = self.bot.max_likes_to_like + 1
        expect_filtered += 1
        my_test_photo_items[3]["like_count"] = self.bot.min_likes_to_like - 1
        expect_filtered += 1
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results,
            "status": "ok",
            "more_available": False,
            "items": my_test_photo_items,
        }

        responses.add(
            responses.GET,
            (
                "{api_url}feed/tag/{hashtag}/?max_id={max_id}&rank_token=" +
                "{rank_token}&ranked_content=true&"
            ).format(
                api_url=API_URL,
                hashtag=hashtag,
                max_id="",
                rank_token=self.bot.api.rank_token,
            ),
            json=response_data,
            status=200,
        )

        medias = self.bot.get_hashtag_medias(hashtag, filtration=False)
        assert medias == [
            test_photo_item["id"] for test_photo_item in my_test_photo_items
        ]
        assert len(medias) == results

        medias = self.bot.get_hashtag_medias(hashtag, filtration=True)
        assert medias == [
            test_photo_item["id"]
            for test_photo_item in my_test_photo_items
            if (
                not test_photo_item["has_liked"]
                and test_photo_item["like_count"] < self.bot.max_likes_to_like
                and test_photo_item["like_count"] > self.bot.min_likes_to_like
            )
        ]
        assert len(medias) == results - expect_filtered

    @responses.activate
    @pytest.mark.parametrize("hashtag", ["hashtag1", "hashtag2"])
    def test_get_total_hashtag_medias(self, hashtag):

        amount = 5
        results = 10
        my_test_photo_item = TEST_PHOTO_ITEM.copy()
        my_test_photo_item["like_count"] = self.bot.min_likes_to_like + 1
        my_test_photo_items = []
        for _ in range(results):
            my_test_photo_items.append(my_test_photo_item.copy())
        expect_filtered = 0
        my_test_photo_items[1]["has_liked"] = True
        expect_filtered += 1
        my_test_photo_items[2]["like_count"] = self.bot.max_likes_to_like + 1
        expect_filtered += 1
        my_test_photo_items[3]["like_count"] = self.bot.min_likes_to_like - 1
        expect_filtered += 1
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results,
            "status": "ok",
            "more_available": True,
            "next_max_id": TEST_PHOTO_ITEM["id"],
            "items": my_test_photo_items,
        }

        responses.add(
            responses.GET,
            (
                "{api_url}feed/tag/{hashtag}/?max_id={max_id}" +
                "&rank_token={rank_token}&ranked_content=true&"
            ).format(
                api_url=API_URL,
                hashtag=hashtag,
                max_id="",
                rank_token=self.bot.api.rank_token,
            ),
            json=response_data,
            status=200,
        )

        medias = self.bot.get_total_hashtag_medias(
            hashtag, amount=amount, filtration=False
        )
        assert medias == [
            test_photo_item["id"]
            for test_photo_item in my_test_photo_items[:amount]
        ]
        assert len(medias) == amount

        medias = self.bot.get_total_hashtag_medias(
            hashtag, amount=amount, filtration=True
        )
        assert medias == [
            test_photo_item["id"]
            for test_photo_item in my_test_photo_items[:amount]
            if (
                not test_photo_item["has_liked"]
                and test_photo_item["like_count"] < self.bot.max_likes_to_like
                and test_photo_item["like_count"] > self.bot.min_likes_to_like
            )
        ]
        assert len(medias) == amount - expect_filtered

    @responses.activate
    @pytest.mark.parametrize("media_id", ["1234567890", 1234567890])
    def test_get_media_likers(self, media_id):
        results = 5
        responses.add(
            responses.GET,
            "{api_url}media/{media_id}/likers/?".format(
                api_url=API_URL, media_id=media_id
            ),
            json={
                "user_count": results,
                "status": "ok",
                "users": [TEST_MEDIA_LIKER for _ in range(results)],
            },
            status=200,
        )

        medias = self.bot.get_media_likers(media_id)

        assert medias == [str(TEST_MEDIA_LIKER["pk"]) for _ in range(results)]
        assert len(medias) == results

    @responses.activate
    @pytest.mark.parametrize("user_id", [19, "19"])
    def test_get_last_user_medias(self, user_id):

        results = 5
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results,
            "status": "ok",
            "more_available": False,
            "items": [TEST_PHOTO_ITEM for _ in range(results)],
        }

        responses.add(
            responses.GET,
            (
                "{api_url}feed/user/{user_id}/?max_id={max_id}&min_timestamp" +
                "={min_timestamp}&rank_token={rank_token}&ranked_content=true"
            ).format(
                api_url=API_URL,
                user_id=user_id,
                max_id="",
                min_timestamp=None,
                rank_token=self.bot.api.rank_token,
            ),
            json=response_data,
            status=200,
        )

        medias = self.bot.get_last_user_medias(user_id, count=results)

        assert medias == [TEST_PHOTO_ITEM["id"] for _ in range(results)]
        assert len(medias) == results

    @responses.activate
    @pytest.mark.parametrize("user_id", [19, "19"])
    def test_get_total_user_medias(self, user_id):

        results = 18
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results,
            "status": "ok",
            "more_available": False,
            "items": [TEST_PHOTO_ITEM for _ in range(results)],
        }

        responses.add(
            responses.GET,
            (
                "{api_url}feed/user/{user_id}/?max_id={max_id}&min_timestamp" +
                "={min_timestamp}&rank_token={rank_token}&ranked_content=true"
            ).format(
                api_url=API_URL,
                user_id=user_id,
                max_id="",
                min_timestamp=None,
                rank_token=self.bot.api.rank_token,
            ),
            json=response_data,
            status=200,
        )

        medias = self.bot.get_total_user_medias(user_id)
        assert medias == [TEST_PHOTO_ITEM["id"] for _ in range(results)]
        assert len(medias) == results

    @responses.activate
    @pytest.mark.parametrize("user_id", ["1234567890", 1234567890])
    def test_get_user_likers(self, user_id):

        results_1 = 1
        responses.add(
            responses.GET,
            (
                "{api_url}feed/user/{user_id}/?max_id={max_id}&min_timestamp" +
                "={min_timestamp}&rank_token={rank_token}&ranked_content=true"
            ).format(
                api_url=API_URL,
                user_id=user_id,
                max_id="",
                min_timestamp=None,
                rank_token=self.bot.api.rank_token,
            ),
            json={
                "auto_load_more_enabled": True,
                "num_results": results_1,
                "status": "ok",
                "more_available": False,
                "items": [TEST_PHOTO_ITEM for _ in range(results_1)],
            },
            status=200,
        )

        results_2 = 3
        responses.add(
            responses.GET,
            "{api_url}media/{media_id}/likers/?".format(
                api_url=API_URL, media_id=TEST_PHOTO_ITEM["id"]
            ),
            json={
                "user_count": results_2,
                "status": "ok",
                "users": [TEST_MEDIA_LIKER for _ in range(results_2)],
            },
            status=200,
        )

        user_ids = self.bot.get_user_likers(user_id)

        assert user_ids == list(
            {str(TEST_MEDIA_LIKER["pk"]) for _ in range(results_2)}
        )
        assert len(user_ids) == len(
            list({str(TEST_MEDIA_LIKER["pk"]) for _ in range(results_2)})
        )

    @responses.activate
    @pytest.mark.parametrize("username", ["1234567890", 1234567890])
    def test_get_user_followers(self, username):

        test_username = "test.username"

        response_data_1 = {"status": "ok", "user": TEST_SEARCH_USERNAME_ITEM}
        responses.add(
            responses.GET,
            "{api_url}users/{username}/usernameinfo/".format(
                api_url=API_URL, username=test_username
            ),
            status=200,
            json=response_data_1,
        )

        response_data_2 = {"status": "ok", "user": TEST_USERNAME_INFO_ITEM}
        responses.add(
            responses.GET,
            "{api_url}users/{user_id}/info/".format(
                api_url=API_URL,
                user_id=username
            ),
            status=200,
            json=response_data_2,
        )

        results_3 = 5
        response_data_3 = {
            "status": "ok",
            "big_list": False,
            "next_max_id": None,
            "sections": None,
            "users": [TEST_FOLLOWER_ITEM for _ in range(results_3)],
        }
        responses.add(
            responses.GET,
            (
                "{api_url}friendships/{user_id}/followers/?" +
                "rank_token={rank_token}"
            ).format(
                api_url=API_URL,
                user_id=username,
                rank_token=self.bot.api.rank_token
            ),
            json=response_data_3,
            status=200,
        )

        user_ids = self.bot.get_user_followers(username)

        assert user_ids == [
            str(TEST_FOLLOWER_ITEM["pk"]) for _ in range(results_3)
        ]

    @responses.activate
    @pytest.mark.parametrize("username", ["1234567890", 1234567890])
    def test_get_user_following(self, username):

        test_username = "test.username"

        response_data_1 = {"status": "ok", "user": TEST_SEARCH_USERNAME_ITEM}
        responses.add(
            responses.GET,
            "{api_url}users/{username}/usernameinfo/".format(
                api_url=API_URL, username=test_username
            ),
            status=200,
            json=response_data_1,
        )

        response_data_2 = {"status": "ok", "user": TEST_USERNAME_INFO_ITEM}
        responses.add(
            responses.GET,
            "{api_url}users/{user_id}/info/".format(
                api_url=API_URL,
                user_id=username
            ),
            status=200,
            json=response_data_2,
        )

        results_3 = 5
        response_data_3 = {
            "status": "ok",
            "big_list": False,
            "next_max_id": None,
            "sections": None,
            "users": [TEST_FOLLOWING_ITEM for _ in range(results_3)],
        }
        responses.add(
            responses.GET,
            (
                "{api_url}friendships/{user_id}/following/?max_id={max_id}" +
                "&ig_sig_key_version={sig_key}&rank_token={rank_token}"
            ).format(
                api_url=API_URL,
                user_id=username,
                rank_token=self.bot.api.rank_token,
                sig_key=SIG_KEY_VERSION,
                max_id="",
            ),
            json=response_data_3,
            status=200,
        )

        user_ids = self.bot.get_user_following(username)

        assert user_ids == [
            str(TEST_FOLLOWING_ITEM["pk"]) for _ in range(results_3)
        ]

    @responses.activate
    @pytest.mark.parametrize("hashtag", ["hashtag1", "hashtag2"])
    def test_get_hashtag_users(self, hashtag):

        results = 5
        my_test_photo_item = TEST_PHOTO_ITEM.copy()
        my_test_photo_item["like_count"] = self.bot.min_likes_to_like + 1
        my_test_photo_items = []
        for _ in range(results):
            my_test_photo_items.append(my_test_photo_item.copy())
        expect_filtered = 0
        my_test_photo_items[1]["has_liked"] = True
        expect_filtered += 1
        my_test_photo_items[2]["like_count"] = self.bot.max_likes_to_like + 1
        expect_filtered += 1
        my_test_photo_items[3]["like_count"] = self.bot.min_likes_to_like - 1
        expect_filtered += 1
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results,
            "status": "ok",
            "more_available": False,
            "items": my_test_photo_items,
        }

        responses.add(
            responses.GET,
            (
                "{api_url}feed/tag/{hashtag}/?max_id={max_id}" +
                "&rank_token={rank_token}&ranked_content=true&"
            ).format(
                api_url=API_URL,
                hashtag=hashtag,
                max_id="",
                rank_token=self.bot.api.rank_token,
            ),
            json=response_data,
            status=200,
        )

        medias = self.bot.get_hashtag_users(hashtag)
        assert medias == [
            str(test_photo_item["user"]["pk"])
            for test_photo_item in my_test_photo_items
        ]
        assert len(medias) == results

    @responses.activate
    @pytest.mark.parametrize(
        "comment_id",
        [
            "12345678901234567",
            12345678901234567
        ]
    )
    def test_get_comment_likers(self, comment_id):
        results = 5
        response_data = {
            "status": "ok",
            "users": [TEST_COMMENT_LIKER_ITEM for _ in range(results)],
        }
        responses.add(
            responses.GET,
            "{api_url}media/{comment_id}/comment_likers/?".format(
                api_url=API_URL, comment_id=comment_id
            ),
            json=response_data,
            status=200,
        )
        user_ids = self.bot.get_comment_likers(comment_id)
        assert user_ids == [
            str(TEST_COMMENT_LIKER_ITEM["pk"])
            for _ in range(results)
        ]
        assert len(user_ids) == results

    @responses.activate
    @pytest.mark.parametrize("latitude", [1.2345])
    @pytest.mark.parametrize("longitude", [9.8765])
    def test_get_locations_from_coordinates(self, latitude, longitude):
        results = 10
        response_data = {
            "has_more": False,
            "items": [TEST_LOCATION_ITEM for _ in range(results)],
            "rank_token": self.bot.api.rank_token,
            "status": "ok",
        }
        responses.add(
            responses.GET,
            (
                "{api_url}fbsearch/places/?rank_token={rank_token}" +
                "&query={query}&lat={lat}&lng={lng}"
            ).format(
                api_url=API_URL,
                rank_token=self.bot.api.rank_token,
                query="",
                lat=latitude,
                lng=longitude,
            ),
            json=response_data,
            status=200,
        )
        locations = self.bot.get_locations_from_coordinates(
            latitude,
            longitude
        )
        assert locations == [TEST_LOCATION_ITEM for _ in range(results)]
        assert len(locations) == results

    @responses.activate
    def test_get_messages(self):
        results = 5
        response_data = {
            u"status": "ok",
            u"pending_requests_total": 2,
            u"seq_id": 182,
            u"snapshot_at_ms": 1547815538244,
            u"most_recent_inviter": TEST_MOST_RECENT_INVITER_ITEM,
            u"inbox": {
                u"blended_inbox_enabled": True,
                u"has_older": False,
                u"unseen_count": 1,
                u"unseen_count_ts": 1547815538242025,
                u"threads": [TEST_INBOX_THREAD_ITEM for _ in range(results)],
            },
        }
        responses.add(
            responses.POST,
            "{api_url}direct_v2/inbox/".format(api_url=API_URL),
            json=response_data,
            status=200,
        )
        inbox = self.bot.get_messages()
        assert inbox == response_data



from tqdm import tqdm


def unlike(self, media_id):
    if not self.reached_limit("unlikes"):
        self.delay("unlike")
        if self.api.unlike(media_id):
            self.total["unlikes"] += 1
            return True
    else:
        self.logger.info("Out of unlikes for today.")
    return False


def unlike_comment(self, comment_id):
    if self.api.unlike_comment(comment_id):
        return True
    return False


def unlike_media_comments(self, media_id):
    broken_items = []
    media_comments = self.get_media_comments(media_id)
    comment_ids = [
        item["pk"] for item in media_comments if item["has_liked_comment"]
    ]

    if not comment_ids:
        self.logger.info(
            "None comments received: comments not found"
            " or comments have been filtered."
        )
        return broken_items

    self.logger.info("Going to unlike %d comments." % (len(comment_ids)))

    for comment in tqdm(comment_ids):
        if not self.unlike_comment(comment):
            self.error_delay()
            broken_items = comment_ids[comment_ids.index(comment):]
    self.logger.info(
        "DONE: Unliked {count} comments.".format(
            count=len(comment_ids) - len(broken_items)
        )
    )
    return broken_items


def unlike_medias(self, medias):
    broken_items = []
    self.logger.info("Going to unlike %d medias." % (len(medias)))
    for media in tqdm(medias):
        if not self.unlike(media):
            self.error_delay()
            broken_items = medias[medias.index(media):]
            break
    self.logger.info("DONE: Total unliked %d medias." % self.total["unlikes"])
    return broken_items


def unlike_user(self, user_id):
    self.logger.info("Going to unlike user %s's feed:" % user_id)
    user_id = self.convert_to_user_id(user_id)
    medias = self.get_user_medias(user_id, filtration=False)
    return self.unlike_medias(medias)



def download_stories(self, username):
    user_id = self.get_user_id_from_username(username)
    list_image, list_video = self.get_user_stories(user_id)
    if list_image == [] and list_video == []:
        self.logger.error(
            (
                "Make sure that '{}' is NOT private and that "
                "posted some stories"
            ).format(
                username
            )
        )
        return False
    self.logger.info("Downloading stories...")
    for story_url in list_image:
        filename = story_url.split("/")[-1].split(".")[0] + ".jpg"
        self.api.download_story(filename, story_url, username)
    for story_url in list_video:
        filename = story_url.split("/")[-1].split(".")[0] + ".mp4"
        self.api.download_story(filename, story_url, username)


def upload_story_photo(self, photo, upload_id=None):
    self.small_delay()
    if self.api.upload_story_photo(photo, upload_id):
        self.logger.info("Photo '{}' is uploaded as Story.".format(photo))
        return True
    self.logger.info("Photo '{}' is not uploaded.".format(photo))
    return False


def watch_users_reels(self, user_ids, max_users=100):
    """
        user_ids - the list of user_id to get their stories
        max_users - max amount of users to get stories from.

        It seems like Instagram doesn't allow to get stories
        from more that 100 users at once.
    """

    # In case of only one user were passed
    if not isinstance(user_ids, list):
        user_ids = [user_ids]

    # Get users reels
    reels = self.api.get_users_reel(user_ids[:max_users])

    # Filter to have users with at least 1 reel
    if isinstance(reels, list):
        # strange output
        return False

    reels = {
        k: v for k,
        v in reels.items() if "items" in v and len(v["items"]) > 0
    }

    # Filter reels that were not seen before
    unseen_reels = []
    for _, reels_data in reels.items():
        last_reel_seen_at = reels_data["seen"] if "seen" in reels_data else 0
        unseen_reels.extend(
            [
                r
                for r in reels_data["items"]
                if r["taken_at"] > last_reel_seen_at
            ]
        )

    # See reels that were not seen before
    # TODO: add counters for watched stories
    if self.api.see_reels(unseen_reels):
        self.total["stories_viewed"] += len(unseen_reels)
        return True
    return False



"""
    Filter functions for media and user lists.
"""


def filter_medias(
    self,
    media_items,
    filtration=True,
    quiet=False,
    is_comment=False
):
    if filtration:
        if not quiet:
            self.logger.info("Received {} medias.".format(len(media_items)))
        if not is_comment:
            media_items = _filter_medias_not_liked(media_items)
            if self.max_likes_to_like:
                media_items = _filter_medias_nlikes(
                    media_items, self.max_likes_to_like, self.min_likes_to_like
                )
        else:
            media_items = _filter_medias_not_commented(self, media_items)
        if not quiet:
            msg = "After filtration {} medias left."
            self.logger.info(msg.format(len(media_items)))
    return _get_media_ids(media_items)


def _filter_medias_not_liked(media_items):
    not_liked_medias = []
    for media in media_items:
        if "has_liked" in media and not media["has_liked"]:
            not_liked_medias.append(media)
    return not_liked_medias


def _filter_medias_not_commented(self, media_items):
    not_commented_medias = []
    for media in media_items:
        if media.get("comment_count", 0) > 0 and media.get("comments"):
            my_comments = [
                comment
                for comment in media["comments"]
                if comment["user_id"] == self.user_id
            ]
            if my_comments:
                continue
        not_commented_medias.append(media)
    return not_commented_medias


def _filter_medias_nlikes(media_items, max_likes_to_like, min_likes_to_like):
    filtered_medias = []
    for media in media_items:
        if "like_count" in media:
            if (
                media["like_count"] < max_likes_to_like
                and media["like_count"] > min_likes_to_like
            ):
                filtered_medias.append(media)
    return filtered_medias


def _get_media_ids(media_items):
    result = []
    for media in media_items:
        if "id" in media:
            result.append(media["id"])
        elif "pk" in media:
            result.append(media["pk"])
    return result


def check_media(self, media_id):
    if self.api.media_info(media_id):
        medias = self.api.last_json["items"]

        if search_blacklist_hashtags_in_media(self, media_id):
            msg = "Blacklist hashtag found in media, skipping!"
            self.console_print(msg, "red")
            return False

        if self.filter_medias(medias, quiet=True):
            return check_user(self, self.get_media_owner(media_id))
        return False

    msg = "Media ID error!"
    self.console_print(msg, "red")
    return False


# Filter users


def search_stop_words_in_user(self, user_info):
    text = ""
    if "biography" in user_info:
        text += user_info["biography"].lower()

    if "username" in user_info:
        text += user_info["username"].lower()

    if "full_name" in user_info:
        text += user_info["full_name"].lower()

    for stop_word in self.stop_words:
        if stop_word in text:
            return True

    return False


def search_blacklist_hashtags_in_media(self, media_id):
    media_info = self.get_media_info(media_id)
    text = media_info[0]["caption"]["text"] if media_info[0]["caption"] else ""

    media_comments = self.get_media_comments(media_id)
    comments_number = min(6, len(media_comments))

    for i in range(0, comments_number):
        text += "".join(media_comments[i]["text"])

    return any((h in text) for h in self.blacklist_hashtags)


def check_user(self, user_id, unfollowing=False):  # noqa: C901
    if not self.filter_users and not unfollowing:
        return True

    self.small_delay()
    user_id = self.convert_to_user_id(user_id)

    if not user_id:
        self.console_print("not user_id, skipping!", "red")
        return False
    if user_id in self.whitelist:
        self.console_print("`user_id` in `self.whitelist`.", "green")
        return True
    if user_id in self.blacklist:
        self.console_print("`user_id` in `self.blacklist`.", "red")
        return False

    if user_id == str(self.user_id):
        self.console_print((
            "`user_id` equals bot's `user_id`, skipping!"
        ), "green")
        return False

    if user_id in self.following:
        if not unfollowing:
            # Log to Console
            self.console_print("Already following, skipping!", "red")
        return False

    user_info = self.get_user_info(user_id)
    if not user_info:
        self.console_print("not `user_info`, skipping!", "red")
        return False

    msg = (
        "USER_NAME: {username}, FOLLOWER: {followers}, FOLLOWING: {following}"
    )
    follower_count = user_info["follower_count"]
    following_count = user_info["following_count"]
    self.console_print(
        msg.format(
            username=user_info["username"],
            followers=follower_count,
            following=following_count,
        )
    )

    skipped = self.skipped_file
    followed = self.followed_file

    if not unfollowing:
        if self.filter_previously_followed and user_id in followed.list:
            self.console_print((
                "info: account previously followed, skipping!"
            ), "red")
            return False
    if (
        "has_anonymous_profile_picture" in user_info
        and self.filter_users_without_profile_photo
    ):
        if user_info["has_anonymous_profile_picture"]:
            self.console_print(
                (
                    "info: account DOES NOT HAVE "
                    "A PROFILE PHOTO, skipping! "
                ), "red"
            )
            skipped.append(user_id)
            return False
    if "is_private" in user_info and self.filter_private_users:
        if user_info["is_private"]:
            self.console_print("info: account is PRIVATE, skipping! ", "red")
            skipped.append(user_id)
            return False
    if "is_business" in user_info and self.filter_business_accounts:
        if user_info["is_business"]:
            self.console_print("info: is BUSINESS, skipping!", "red")
            skipped.append(user_id)
            return False
    if "is_verified" in user_info and self.filter_verified_accounts:
        if user_info["is_verified"]:
            self.console_print("info: is VERIFIED, skipping !", "red")
            skipped.append(user_id)
            return False

    if follower_count < self.min_followers_to_follow:
        msg = "follower_count < bot.min_followers_to_follow, skipping!"
        self.console_print(msg, "red")
        skipped.append(user_id)
        return False
    if follower_count > self.max_followers_to_follow:
        msg = "follower_count > bot.max_followers_to_follow, skipping!"
        self.console_print(msg, "red")
        skipped.append(user_id)
        return False
    if user_info["following_count"] < self.min_following_to_follow:
        msg = "following_count < bot.min_following_to_follow, skipping!"
        self.console_print(msg, "red")
        skipped.append(user_id)
        return False
    if user_info["following_count"] > self.max_following_to_follow:
        msg = "following_count > bot.max_following_to_follow, skipping!"
        self.console_print(msg, "red")
        skipped.append(user_id)
        return False
    try:
        if (
            (following_count > 0)
            and follower_count / following_count >
            self.max_followers_to_following_ratio
        ):
            msg = (
                "follower_count / following_count > "
                "bot.max_followers_to_following_ratio, skipping!"
            )
            self.console_print(msg, "red")
            skipped.append(user_id)
            return False
        if (
            (follower_count > 0) and
            following_count / follower_count >
            self.max_following_to_followers_ratio
        ):
            msg = (
                "following_count / follower_count > "
                "bot.max_following_to_followers_ratio, skipping!"
            )
            self.console_print(msg, "red")
            skipped.append(user_id)
            return False
    except ZeroDivisionError:
        self.console_print("ZeroDivisionError: division by zero", "red")
        return False

    if (
        "media_count" in user_info
        and user_info["media_count"] < self.min_media_count_to_follow
    ):
        msg = (
            "media_count < bot.min_media_count_to_follow, "
            "BOT or INACTIVE, skipping!"
        )
        self.console_print(msg, "red")
        skipped.append(user_id)
        return False

    if search_stop_words_in_user(self, user_info):
        msg = "`bot.search_stop_words_in_user` found in user, skipping!"
        self.console_print(msg, "red")
        skipped.append(user_id)
        return False

    return True


def check_not_bot(self, user_id):
    """ Filter bot from real users. """
    self.small_delay()
    user_id = self.convert_to_user_id(user_id)
    if not user_id:
        return False
    if user_id in self.whitelist:
        return True
    if user_id in self.blacklist:
        return False

    user_info = self.get_user_info(user_id)
    if not user_info:
        return True  # closed acc

    skipped = self.skipped_file
    if (
        "following_count" in user_info
        and user_info["following_count"] > self.max_following_to_block
    ):
        msg = "following_count > bot.max_following_to_block, skipping!"
        self.console_print(msg, "red")
        skipped.append(user_id)
        return False  # massfollower

    if search_stop_words_in_user(self, user_info):
        msg = "`bot.search_stop_words_in_user` found in user, skipping!"
        skipped.append(user_id)
        return False

    return True



import pytest
import responses

from instabot.api.config import API_URL

from .test_bot import TestBot
from .test_variables import (
    TEST_CAPTION_ITEM,
    TEST_COMMENT_ITEM,
    TEST_PHOTO_ITEM,
    TEST_USERNAME_INFO_ITEM,
)

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


class TestBotFilter(TestBot):
    @responses.activate
    @pytest.mark.parametrize(
        "media_id,total,max_per_day",
        [
            [111111, 1, 2],
            [111111, 2, 2],
            [111111, 3, 2],
            ["111111", 1, 2],
            ["111111", 2, 2],
            ["111111", 3, 2],
        ],
    )
    @patch("time.sleep", return_value=None)
    def test_unlike(self, patched_time_sleep, media_id, total, max_per_day):
        self.bot.total["unlikes"] = total
        self.bot.max_per_day["unlikes"] = max_per_day
        responses.add(
            responses.POST,
            "{api_url}media/{media_id}/unlike/".format(
                api_url=API_URL, media_id=media_id
            ),
            json="{'status': 'ok'}",
            status=200,
        )
        _r = self.bot.unlike(media_id)
        test_r = _r if total < max_per_day else not _r
        test_unliked = (
            self.bot.total["unlikes"] == total + 1
            if total < max_per_day
            else self.bot.total["unlikes"] == total
        )
        assert test_r and test_unliked

    @responses.activate
    @pytest.mark.parametrize("comment_id", [111111, "111111"])
    @patch("time.sleep", return_value=None)
    def test_unlike_comment(self, patched_time_sleep, comment_id):
        responses.add(
            responses.POST,
            "{api_url}media/{comment_id}/comment_unlike/".format(
                api_url=API_URL, comment_id=comment_id
            ),
            json="{'status': 'ok'}",
            status=200,
        )
        _r = self.bot.unlike_comment(comment_id)
        assert _r

    @responses.activate
    @pytest.mark.parametrize(
        "media_ids,total,max_per_day",
        [
            [[111111, 222222], 1, 3],
            [[111111, 222222], 2, 3],
            [[111111, 222222], 3, 3],
            [["111111", "222222"], 1, 3],
            [["111111", "222222"], 2, 3],
            [["111111", "222222"], 3, 3],
        ],
    )
    @patch("time.sleep", return_value=None)
    def test_unlike_medias(
        self,
        patched_time_sleep,
        media_ids,
        total,
        max_per_day
    ):
        self.bot.total["unlikes"] = total
        self.bot.max_per_day["unlikes"] = max_per_day
        for media_id in media_ids:
            responses.add(
                responses.POST,
                "{api_url}media/{media_id}/unlike/".format(
                    api_url=API_URL, media_id=media_id
                ),
                json="{'status': 'ok'}",
                status=200,
            )
        broken_items = self.bot.unlike_medias(media_ids)
        test_unliked = self.bot.total["unlikes"] == max_per_day
        test_broken = len(broken_items) == \
            len(media_ids) - (max_per_day - total)
        assert test_unliked and test_broken

    @responses.activate
    @patch("time.sleep", return_value=None)
    def test_unlike_media_comments(self, patched_time_sleep):
        my_test_comment_items = []
        results = 5
        for i in range(results):
            my_test_comment_items.append(TEST_COMMENT_ITEM.copy())
            my_test_comment_items[i]["pk"] = TEST_COMMENT_ITEM["pk"] + i
            if i % 2:
                my_test_comment_items[i]["has_liked_comment"] = False
            else:
                my_test_comment_items[i]["has_liked_comment"] = True
        media_id = 1234567890
        response_data = {
            "caption": TEST_CAPTION_ITEM,
            "caption_is_edited": False,
            "comment_count": results,
            "comment_likes_enabled": True,
            "comments": my_test_comment_items,
            "has_more_comments": False,
            "has_more_headload_comments": False,
            "media_header_display": "none",
            "preview_comments": [],
            "status": "ok",
        }
        responses.add(
            responses.GET,
            "{api_url}media/{media_id}/comments/?".format(
                api_url=API_URL, media_id=media_id
            ),
            json=response_data,
            status=200,
        )
        for my_test_comment_item in my_test_comment_items:
            responses.add(
                responses.POST,
                "{api_url}media/{comment_id}/comment_unlike/".format(
                    api_url=API_URL, comment_id=my_test_comment_item["pk"]
                ),
                json="{'status': 'ok'}",
                status=200,
            )
        broken_items = self.bot.unlike_media_comments(media_id)
        assert broken_items == []

    @responses.activate
    @patch("time.sleep", return_value=None)
    def test_unlike_user(self, patched_time_sleep):
        unliked_at_start = self.bot.total["unlikes"]
        user_id = 1234567890
        response_data = {"status": "ok", "user": TEST_USERNAME_INFO_ITEM}
        responses.add(
            responses.GET,
            "{api_url}users/{username}/usernameinfo/".format(
                api_url=API_URL, username=user_id
            ),
            status=200,
            json=response_data,
        )
        my_test_photo_items = []
        results = 5
        for i in range(results):
            my_test_photo_items.append(TEST_PHOTO_ITEM.copy())
            my_test_photo_items[i]["pk"] = TEST_PHOTO_ITEM["id"] + i
            if i % 2:
                my_test_photo_items[i]["has_liked"] = False
            else:
                my_test_photo_items[i]["has_liked"] = True
        response_data = {
            "auto_load_more_enabled": True,
            "num_results": results,
            "status": "ok",
            "more_available": False,
            "items": my_test_photo_items,
        }
        responses.add(
            responses.GET,
            "{api_url}feed/user/{user_id}/".format(
                api_url=API_URL,
                user_id=user_id
            ),
            json=response_data,
            status=200,
        )
        for my_test_photo_item in my_test_photo_items:
            responses.add(
                responses.POST,
                "{api_url}media/{media_id}/unlike/".format(
                    api_url=API_URL, media_id=my_test_photo_item["id"]
                ),
                json="{'status': 'ok'}",
                status=200,
            )
        broken_items = self.bot.unlike_user(user_id)
        test_unliked = self.bot.total["unlikes"] == unliked_at_start + len(
            my_test_photo_items
        )
        test_broken = broken_items == []
        assert test_broken and test_unliked



import os
from io import open

from tqdm import tqdm


def upload_photo(
    self,
    photo,
    caption=None,
    upload_id=None,
    from_video=False,
    options={}
):
    """Upload photo to Instagram

    @param photo       Path to photo file (String)
    @param caption     Media description (String)
    @param upload_id   Unique upload_id (String). When None, then
                       generate automatically
    @param from_video  A flag that signals whether the photo is loaded from
                       the video or by itself (Boolean, DEPRECATED: not used)
    @param options     Object with difference options, e.g.
                       configure_timeout, rename (Dict)
                       Designed to reduce the number of function arguments!
                       This is the simplest request object.

    @return            Object with state of uploading to Instagram (or False)
    """
    self.small_delay()
    result = self.api.upload_photo(
        photo,
        caption,
        upload_id,
        from_video,
        options=options
    )
    if not result:
        self.logger.info("Photo '{}' is not uploaded.".format(photo))
        return False
    self.logger.info("Photo '{}' is uploaded.".format(photo))
    return result


def download_photo(
    self,
    media_id,
    folder="photos",
    filename=None,
    save_description=False
):
    self.small_delay()

    if not os.path.exists(folder):
        os.makedirs(folder)

    if save_description:
        media = self.get_media_info(media_id)[0]
        caption = media["caption"]["text"] if media["caption"] else ""
        username = media["user"]["username"]
        fname = os.path.join(folder, "{}_{}.txt".format(username, media_id))
        with open(fname, encoding="utf8", mode="w") as f:
            f.write(caption)

    try:
        return self.api.download_photo(media_id, filename, False, folder)
    except Exception:
        self.logger.info("Media with `{}` is not downloaded.".format(media_id))
        return False


def download_photos(self, medias, folder, save_description=False):
    broken_items = []

    if not medias:
        self.logger.info("Nothing to downloads.")
        return broken_items

    self.logger.info("Going to download {} medias.".format(len(medias)))

    for media in tqdm(medias):
        if not self.download_photo(
            media, folder, save_description=save_description
        ):
            self.error_delay()
            broken_items = medias[medias.index(media):]
    return broken_items
